<?php

/**
 * @file
 * Drush interface to image optimize effect.
 */

/**
 * Implements hook_drush_command().
 */
function image_optimize_effect_drush_command() {
  $commands = array();
  $commands['image-optimize-effect-all'] = array(
    'description' => 'Optimize all generated images.',
    'callback' => 'drush_image_optimize_effect_all',
    'aliases' => array('ioea', 'optimize-images-all'),
    'bootstrap' => DRUSH_BOOTSTRAP_DRUPAL_FULL,
    'arguments' => array(
      'limit' => dt("The number of images to convert. Set to 0 to process all images. Defaults to 0."),
      'speed' => dt('Speed/quality trade-off from 1 (brute-force) to 10 (fastest). The default is 3. Speed 10 has 5% lower quality, but is 8 times faster than the default.'),
    ),
    'options' => array(
      'no-interlace' => 'Do not interlace jpg.',
    ),
  );
  $commands['image-optimize-effect-png'] = array(
    'description' => 'Optimize all generated png images.',
    'callback' => 'drush_image_optimize_effect_png',
    'aliases' => array('ioep', 'optimize-images-png'),
    'bootstrap' => DRUSH_BOOTSTRAP_DRUPAL_FULL,
    'arguments' => array(
      'limit' => dt("The number of images to convert. Set to 0 to process all images. Defaults to 0."),
      'speed' => dt('Speed/quality trade-off from 1 (brute-force) to 10 (fastest). The default is 3. Speed 10 has 5% lower quality, but is 8 times faster than the default.'),
    ),
  );
  $commands['image-optimize-effect-jpg'] = array(
    'description' => 'Optimize all generated jpeg images.',
    'callback' => 'drush_image_optimize_effect_jpg',
    'aliases' => array('ioej', 'optimize-images-jpg'),
    'bootstrap' => DRUSH_BOOTSTRAP_DRUPAL_FULL,
    'arguments' => array(
      'limit' => dt("The number of images to convert. Set to 0 to process all images. Defaults to 0."),
    ),
    'options' => array(
      'no-interlace' => 'Do not interlace jpg.',
    ),
  );
  return $commands;
}

/**
 * Callback for command image-optimize-effect-all.
 */
function drush_image_optimize_effect_all($limit = NULL, $speed = 3) {
  drush_image_optimize_effect_png($limit, $speed);
  drush_image_optimize_effect_jpg($limit);
}

/**
 * Callback for command image-optimize-effect-png.
 */
function drush_image_optimize_effect_png($limit = NULL, $speed = 3) {
  $last_run = variable_get('image_optimize_effect_last_run', 0);
  $counter = 0;
  foreach(file_scan_directory('public://styles/', '/.png/') as $file) {
    if (!is_writable($file->uri)) {
      drush_log(dt('@file is not writable.', array(
        '@file' => $file->uri,
      )), 'ok');
    }
    else {
      $file_changed = filemtime($file->uri);
      if ($file_changed > $last_run) {
        $image = image_load($file->uri);
        drush_log(dt('Processing @file, current size is @size.', array(
          '@file' => $file->uri,
          '@size' => $image->info['file_size'],
        )), 'ok');
        if (image_optimize_effect_pngquant($image, array('speed' => $speed))) {
          // image is already saved, do not use image_save($image);
        }
        drush_log(dt('Done processing @file, new size is @size.', array(
          '@file' => $file->uri,
          '@size' => $image->info['file_size'],
        )), 'ok');
        imagedestroy($image->resource);

        // Check limit if needed.
        $counter++;
        if (isset($limit) && $counter >= $limit) {
          return;
        }
      }
    }
  }
  variable_set('image_optimize_effect_last_run', time());
}

/**
 * Callback for command image-optimize-effect-jpg.
 */
function drush_image_optimize_effect_jpg($limit = NULL) {
  $last_run = variable_get('image_optimize_effect_last_run', 0);
  // Recurse all files in styles.
  foreach(file_scan_directory('public://styles/', '/.jpg|.jpeg/') as $file) {
    if (!is_writable($file->uri)) {
      drush_log(dt('@file is not writable.', array(
        '@file' => $file->uri,
      )), 'ok');
    }
    else {
      $file_changed = filemtime($file->uri);
      if ($file_changed > $last_run) {
        $image = image_load($file->uri);
        drush_log(dt('Processing @file, current size is @size.', array(
          '@file' => $file->uri,
          '@size' => $image->info['file_size'],
        )), 'ok');
        if (image_optimize_effect_imgmin($image)) {
          // image is already saved, do not use image_save($image);
        }
        drush_log(dt('Done processing @file, new size is @size.', array(
          '@file' => $file->uri,
          '@size' => filesize($file->uri),
        )), 'ok');
        imagedestroy($image->resource);

        // Check limit if needed.
        $counter++;
        if (isset($limit) && $counter >= $limit) {
          return;
        }
      }
    }
  }
  variable_set('image_optimize_effect_last_run', time());
}

/**
 * Image effect callback; Optimize using pngquant.
 *
 * @param $image
 *   An image object returned by image_load().
 *
 * @return
 *   TRUE on success. FALSE on failure.
 *
 */
function image_optimize_effect_pngquant($image, $data) {
  $source = drupal_realpath($image->source);
  // Merge defaults
  $data += array(
    'speed' => 10,
  );

  $extension = str_replace('jpg', 'jpeg', $image->info['extension']);

  // Only works on png files.
  if ($extension != 'png') {
    return TRUE;
  }

  // Make a copy of the image and save it.
  $res = image_gd_create_tmp($image, $image->info['width'], $image->info['height']);
  $destination = drupal_tempnam('temporary://', 'gd_') . '.' . $extension;
  $destination = drupal_realpath($destination);
  $function = 'image' . $extension;
  if (!function_exists($function)) {
    return FALSE;
  }

  imagealphablending($image->resource, FALSE);
  imagesavealpha($image->resource, TRUE);
  $success = $function($image->resource, $destination);

  // Use Drupal's root as working directory to resolve relative paths correctly.
  $drupal_path = DRUPAL_ROOT;
  $pngquant_path = '';
  $convert_path = 'pngquant';
  $command_args = '--speed ' . $data['speed'] . ' --ext .png -f -v ' . $destination;

  if (strstr($_SERVER['SERVER_SOFTWARE'], 'Win32') || strstr($_SERVER['SERVER_SOFTWARE'], 'IIS')) {
    // Use Window's start command with the /B flag to make the process run in
    // the background and avoid a shell command line window from showing up.
    // @see http://us3.php.net/manual/en/function.exec.php#56599
    // Use /D to run the command from PHP's current working directory so the
    // file paths don't have to be absolute.
    $convert_path = 'start "pngquant" /D ' . escapeshellarg($drupal_path) . ' /B ' . escapeshellarg($pngquant_path);
  }
  $command = $convert_path . ' ' . $command_args;

  $descriptors = array(
    // stdin
    0 => array('pipe', 'r'),
    // stdout
    1 => array('pipe', 'w'),
    // stderr
    2 => array('pipe', 'w'),
  );
  if ($h = proc_open($command, $descriptors, $pipes, $drupal_path)) {
    $output = '';
    while (!feof($pipes[1])) {
      $output .= fgets($pipes[1]);
    }
    $error = '';
    while (!feof($pipes[2])) {
      $error .= fgets($pipes[2]);
    }

    fclose($pipes[0]);
    fclose($pipes[1]);
    fclose($pipes[2]);
    $return_code = proc_close($h);

    // If ImageMagick returned a non-zero code, trigger a PHP error that will
    // be caught by Drupal's error handler, logged to the watchdog and
    // eventually displayed to the user if configured to do so.
    if ($return_code != 0) {
      // If there is no error message, clarify this.
      if ($error === '') {
        $error = t('No error message.');
      }
      // Format $error with as full message, passed by reference.
      $error = t('ImageMagick error @code: !error', array(
        '@code' => $return_code,
        '!error' => $error,
      ));
      // @todo Use watchdog() instead? Would hide errors from users during
      //   normal operation, regeardless of error_level setting.
      trigger_error($error, E_USER_ERROR);
      // ImageMagick exited with an error code, return it.
      return $return_code;
    }
    // The shell command was executed successfully.

    imagedestroy($image->resource);
    $function = 'imagecreatefrom' . $extension;
    $image->resource = $function($destination);

    // Copy image to original location.
    clearstatcache();
    $image->info = image_get_info($destination, $image->toolkit);
    rename($destination, $source);

    return TRUE;
  }
  // The shell command could not be executed.
  return FALSE;
}

/**
 * Image effect callback; Optimize using imgmin.
 *
 * @param $image
 *   An image object returned by image_load().
 *
 * @return
 *   TRUE on success. FALSE on failure.
 *
 */
function image_optimize_effect_imgmin($image) {
  $extension = str_replace('jpg', 'jpeg', $image->info['extension']);
  $source = drupal_realpath($image->source);

  // Make a copy of the image and save it.
  $res = image_gd_create_tmp($image, $image->info['width'], $image->info['height']);
  $destination = drupal_tempnam('temporary://', 'gd_') . '.' . $extension;
  $destination = drupal_realpath($destination);
  $function = 'image' . $extension;
  if (!function_exists($function)) {
    return FALSE;
  }

  imagealphablending($image->resource, FALSE);
  imagesavealpha($image->resource, TRUE);

  // Check for interlace.
  if (!drush_get_option('no-interlace')) {
    image_optimize_effect_interlace_on($image);
  }

  $success = $function($image->resource, $destination);

  // Use Drupal's root as working directory to resolve relative paths correctly.
  $drupal_path = DRUPAL_ROOT;
  $imgmin_path = '';
  $convert_path = 'imgmin';
  $command_args = ' --quality-in-min 50 ' . $destination . ' ' . $destination;

  if (strstr($_SERVER['SERVER_SOFTWARE'], 'Win32') || strstr($_SERVER['SERVER_SOFTWARE'], 'IIS')) {
    // Use Window's start command with the /B flag to make the process run in
    // the background and avoid a shell command line window from showing up.
    // @see http://us3.php.net/manual/en/function.exec.php#56599
    // Use /D to run the command from PHP's current working directory so the
    // file paths don't have to be absolute.
    $convert_path = 'start "imgmin" /D ' . escapeshellarg($drupal_path) . ' /B ' . escapeshellarg($imgmin_path);
  }
  $command = $convert_path . ' ' . $command_args;

  $descriptors = array(
    // stdin
    0 => array('pipe', 'r'),
    // stdout
    1 => array('pipe', 'w'),
    // stderr
    2 => array('pipe', 'w'),
  );
  if ($h = proc_open($command, $descriptors, $pipes, $drupal_path)) {
    $output = '';
    while (!feof($pipes[1])) {
      $output .= fgets($pipes[1]);
    }
    $error = '';
    while (!feof($pipes[2])) {
      $error .= fgets($pipes[2]);
    }

    fclose($pipes[0]);
    fclose($pipes[1]);
    fclose($pipes[2]);
    $return_code = proc_close($h);

    // If imgmin returned a non-zero code, trigger a PHP error that will
    // be caught by Drupal's error handler, logged to the watchdog and
    // eventually displayed to the user if configured to do so.
    if ($return_code != 0) {
      // If there is no error message, clarify this.
      if ($error === '') {
        $error = t('No error message.');
      }
      // Format $error with as full message, passed by reference.
      $error = t('imgmin error @code: !error', array(
        '@code' => $return_code,
        '!error' => $error,
      ));
      // @todo Use watchdog() instead? Would hide errors from users during
      //   normal operation, regeardless of error_level setting.
      trigger_error($error, E_USER_ERROR);
      // imgmin exited with an error code, return it.
      return $return_code;
    }
    // The shell command was executed successfully.

    imagedestroy($image->resource);
    $function = 'imagecreatefrom' . $extension;
    $image->resource = $function($destination);

    // Copy image to original location.
    clearstatcache();
    $image->info = image_get_info($destination, $image->toolkit);
    rename($destination, $source);

    return TRUE;
  }
  // The shell command could not be executed.
  return FALSE;
}

function image_optimize_effect_interlace_on($image) {
  imageinterlace($image->resource, 1);
  return TRUE;
}
